;
; Project "OS"
;
; Copyright (c) 2001 Sergey Gotsuljak
; Volgograd State Technical University       
;
; Boot    
;                                       
                                          
                              
;         
; This boot code founded on next source:
; V. Anoshin "What is it in the boot sector?", Monitor N2, 1993, p.60-62
; I'm rewrite all in my style of programming
;                                                           
                                              
                .model  tiny
                                       
OSLoaderSegment equ     0100h
OSLoaderOffset  equ     0000h          
                                       
                ;Boot Sector                  
BootSector      struc                    
OEM             db      'SerGo   '     ; OEM ID
BytesPerSector  dw      512            ; Bytes per sector
ClusterSize     db      1              ; Sectors per cluster   
ReservedSectors dw      1              ; Reserved sectors at beginning
FATCopies       db      2              ; Fat copies            
RootSize        dw      224            ; Root directory entries
TotalSectors    dw      2880           ; Total sectors on disk 
Media           db      0F0h           ; Media descriptor byte 
FATSize         dw      9              ; Sectors per FAT       
TrackSize       dw      18             ; Sectors per track  
Heads           dw      2              ; Heads              
HiddenSectors   dd      0              ; Special hidden sectors
BigTotal        dd      0              ; Big total number of sectors
Drive           db      0              ; Physical drive number
q               db      0                               
ExtendedBPB     db      29h            ; Extended boot record signature
SerialNumber    dd      0              ; Volume serial number
VLabel          db      'OS Disk    '  ; Volume label
FSID            db      'FAT12   '     ; File system ID                           
BootSector      ends                     
                                                              
                .data                         
BootName        db      'LOADER  EXE',0  
BootMessage     db      'Boot:',0
NoBootMessage   db      'no loader!',0
ErrorMessage    db      'disk error!',0       
SuccessMessage  db      'OK!',0
RootCount       equ     [bp+4]           
RootBegin       equ     [bp+6]           
Cluster         equ     [bp+8]           
FAT             equ     [bp+10]          
SectorLong      equ     [bp+14]          
Buffer          equ     [bp+18]          
Sector          equ     [bp+22]     
Track           equ     [bp+24]           
Head            equ     [bp+26]
                                                            
; Use only with FAT12!                                                       
_mul            db      3                                                   
EOF             dw      0FF8h                                               
                                                                            
                .code                                                       
                org     100h                                                
                                                                            
main            proc                                                        
                                                                            
                jmp     Start                                               
                nop                                                         
                                                                            
BS              BootSector <>                                               
                                                                            
Start:          ; jmp far 07B0h:Main                                         
                db      0EAh                                                
                dw      offset Main,07B0h                                   
                                                                            
Main:           ; Set registers 
                xor     ax,ax                                               
                mov     ss,ax                                               
                mov     sp,7BD0h
                mov     bp,sp 
                push    cs               
                pop     ds        
                                  
                lea     si,BootMessage
                call    PutStr                      
                                         
                mov     ax,BS.RootSize
                mov     cl,4             
                shr     ax,cl            
                mov     RootCount,ax
                              
                mov     al,BS.FATCopies
                xor     ah,ah         
                mul     BS.FATSize
                add     ax,BS.ReservedSectors
                mov     RootBegin,ax
                          
                ; Read root directory      
                xor     ax,ax           
                mov     word ptr Buffer+2,0070h
                mov     word ptr Buffer,ax
                                                            
@6:             push    ax
                          
                add     ax,RootBegin
                xor     dx,dx         
                mov     cx,1                       
                          
                ; Read next sector of root directory
                call    ReadLS
                sub     word ptr Buffer,512
                  
                ; Search name of boot file
                xor     ax,ax                
@7:                                   
                mov     cx,11         
                les     di,Buffer
                add     di,ax         
                lea     si,BootName
                cld                      
                repe    cmpsb                                               
                jcxz    @8            
                add     ax,32         
                cmp     ax,512        
                jb      @7                                                  
                  
                pop     ax            
                inc     ax            
                cmp     ax,RootCount   
                jb      @6            
                lea     si,NoBootMessage
                call    PutStr                      
@14:                                  
                jmp     @14           
@8:                                   
                pop     bx       
                                 
                les     bx,Buffer
                add     bx,ax                                               
                mov     ax,word ptr es:[bx+26]
                mov     Cluster,ax
                               
                ; Read FAT                 
                mov     dx,0800h        
                xor     ax,ax         
                mov     word ptr FAT+2,dx
                mov     word ptr FAT,ax                     
                mov     word ptr Buffer+2,dx
                mov     word ptr Buffer,ax
                               
                mov     ax,BS.ReservedSectors
                xor     dx,dx         
                mov     cx,BS.FATSize
                                
                call    ReadLS  
                                
                ; Read OS Loader
                mov     word ptr Buffer+2,OSLoaderSegment
                mov     word ptr Buffer,OSLoaderOffset            
@9:                                                   
                mov     ax,Cluster                    
                cmp     ax,EOF                        
                jae     @13                           
                call    ReadCluster                   
                                                      
                ; Compute number of next cluster      
                mov     ax,Cluster                    
                mul     byte ptr _mul                 
                shr     ax,1                          
                les     si,FAT                        
                add     si,ax                         
                mov     ax,es:[si]                    
                                                      
                cmp     byte ptr _mul,3               
                jne     @12                           
                mov     cl,4                          
                test    byte ptr Cluster,1            
                jne     @11                           
                shl     ax,cl                         
@11:                                                  
                shr     ax,cl                                               
@12:                                                  
                mov     Cluster,ax                    
                                                      
                jmp     @9                            
@13:                                                  
                lea     si,SuccessMessage             
                call    PutStr                        
                                                      
                ; Goto boot entry                                            
                db      0EAh                          
                dw      OSLoaderOffset,OSLoaderSegment+0020h
main            endp                                  
                                                                            
; Read logical sector                                 
; dx:ax - number of logical sector                    
ReadLS          proc                                  
                mov     word ptr SectorLong,ax        
                mov     word ptr SectorLong+2,dx      
@47:                                                  
                push    cx                            
                mov     cx,3                          
@48:                                                  
                push    cx                            
                                                      
                mov     ax,word ptr SectorLong        
                mov     dx,word ptr SectorLong+2      
                add     ax,word ptr BS.HiddenSectors  
                adc     dx,word ptr BS.HiddenSectors+2
                div     BS.TrackSize                
                mov     bx,ax         
                inc     dl            
                mov     Sector,dl
                             
                xor     dx,dx         
                div     BS.Heads
                mov     Track,ax
                
                mov     ch,al     
                shr     ax,1      
                shr     ax,1      
                mov     cl,al     
                and     cl,0C0h   
                add     cl,Sector
                                 
                mov     ax,Track
                mul     BS.Heads
                sub     bx,ax
                
                mov     dh,bl           ; Head
                mov     dl,BS.Drive     ; Drive
                mov     ax,0201h        ; Read one sector
                les     bx,Buffer       ; Buffer
                int     13h             
                                                                            
                pop     cx  
                jnc     short @51
                loop    @48 
@49:                      
                lea     si,ErrorMessage
                call    PutStr                      
@50:                        
                jmp     @50                                                 
@51:
                inc     word ptr SectorLong
                adc     word ptr SectorLong+2,0
                add     word ptr Buffer,512
                pop     cx                
                loop    @47               
                ret                       
ReadLS          endp
                                          
; Read cluster
; ax - cluster number
ReadCluster     proc                      
                sub     ax,2              
                mov     dl,BS.ClusterSize
                xor     dh,dh             
                mul     dx                
                add     ax,RootBegin
                adc     dx,0              
                add     ax,RootCount      
                adc     dx,0              
                mov     cl,BS.ClusterSize
                xor     ch,ch
                call    ReadLS      
                ret                       
ReadCluster     endp
                                          
PutStr          proc
                lodsb                   
                cmp     al,00h          
                je      M8                               
                mov     bx,7                             
                mov     ah,0Eh          ;PutChar via BIOS
                int     10h             
                jmp     short PutStr
M8:                                                                         
                ret                                                         
PutStr          endp                                                        
                                                                            
                end     main                                                